/***************************************************************************
                          optimizer.cpp  -  description
                             -------------------
    begin                : Sat Jul 28 2001
    copyright            : (C) 2001 by Chris Wallace
    email                : ranchhard@netflash.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "optimizer.h"
#include <stdio.h>
#include <iostream.h>

Optimizer::Optimizer()
{
	mDoc = 0;
	mTypewriter = mBold = mItalic = mStrong = 0;
	mEmphasis = false;
	//ignoreBold = false;
}

Optimizer::Optimizer(const Optimizer & opt)
{
	operator = (opt);
}

Optimizer::~Optimizer()
{
}

void Optimizer::operator = (const Optimizer & opt)
{

}

bool Optimizer::optimize(void)
{
	while(1) {
		if(!parseDocument())
			break;
	}
	return true;
}

bool Optimizer::parseDocument(void)
{
	mDoc++;
	if(!parseText())
		return false;
	mDoc--;
	return true;
}

bool Optimizer::parseText(void)
{
	char ch;
	char pch = EOF;
	while(1) {
		ch = getchar();
		if(ch == EOF) {
			return false;
		}
		if(ch == '<') {
			 return parseTag();
		}
		else if(!(mTypewriter <= mDoc && mTypewriter > 0)) {
			if(!(isSpaceCharacter(pch) && isSpaceCharacter(ch))) {
				putchar(ch);
			}
		}
		else {
			putchar(ch);
		}
	
		pch = ch;
	}
	return true;
}

bool Optimizer::parseTag(void)
{
	int tagi = 0;
	char mych = '\0';
	char tag[4];
		
	tag[0] = tag[1] = tag[2] = tag[3] = '\0';
	
	while(1) {
		mych = getchar();
		if(mych == EOF) {
			return false;
		}
		if(mych == '>') {
			break;
		}
			
		if(!isSpaceCharacter(mych) && mych != '<') {
			tag[tagi] = mych;
			tagi++;
		}
		if(tagi > 3) {
			cout << "Tagi is greater than 3!" << endl;
			return false;
		}
	}
	
	bool inEscape = false;
	if(tag[0] == '/') {
		inEscape = true;
		tag[0] = tag[1];
		tag[1] = tag[2];
	}
		
	TagType tt = getTagType(tag[0], tag[1]);
	
	if(!inEscape) {
		if(tt == Emphasis && !mStrong) {
			if(emElimination(inEscape))
				return true;
		}
		printTag(tt, false);
		if(!parseDocument()) {
			return false;
		}
	}
	else {
		if(tt == Emphasis && !mStrong) {
			if(emElimination(inEscape))
				return true;
			else {
				/*if(lookahead(tt)) {
					if(!parseDocument())
						return false;
				}*/
				mDoc--;
				printTag(tt, true);
				mDoc++;
			}
		}
		else {
			/*if(lookahead(tt)) {
				if(!parseDocument())
					return false;
			}*/
			mDoc--;
			printTag(tt, true);
			mDoc++;
		}
		/*else {
			mDoc--;
			printTag(tt, true);
			mDoc++;
		}*/
	}
	
	return true;
}

bool Optimizer::isSpaceCharacter(char c)
{
	switch(c) {
		case ' ':
		case '\t':
		case '\n':
		case '\0':
			return true;
	}
	return false;
}

bool Optimizer::lookahead(const TagType & tt)
{
	std::stack<char> savechars;
	char ch;
	bool found = false;
	bool havetag = false;
	while(1) {
		ch = getchar();
		if(ch == EOF)
			break;
		if(ch == '<') {
			havetag = true;
			savechars.push(ch);
			ungetc(ch, stdin);
			break;
		}
		savechars.push(ch);
	}
	cerr << "Have Tag ? " << havetag << endl;
	if(havetag) {
		while(1) {
			ch = getchar();
			if(ch == EOF)
				break;
			if(ch == '<') {
				savechars.push(ch);
				ch = getchar();
				if(ch != '/') {
					//cerr << endl << ch <<endl;
					TagType tagtype = getTagType(ch, '_');
					//cerr << endl << tagtype << " " << tt << endl;
					if(tagtype == tt) {
						found = true;
						savechars.push(ch);
						break;
					}
				}
				else {
					TagType tagtype = getTagType(ch, '_');
					//cerr << endl << tagtype << " " << tt << endl;
					if(tagtype == tt) {
						found = true;
						savechars.push(ch);
						break;
					}
					//break;
				}
			}
			if(!isSpaceCharacter(ch)) {
				found = false;
				savechars.push(ch);
				//break;
			}
			savechars.push(ch);
		}
	}
	while(!savechars.empty()) {
		ch = savechars.top();
		ungetc(ch, stdin);
		savechars.pop();
	}
	//cerr << endl << "Found tag? " << found << " Looking for " << tt << endl;
	return found;
}

bool Optimizer::lookahead(const TagType & tt, bool escape)
{
	std::stack<char> savechars;
	char ch;
	bool found = false;
	bool havetag = false;
	while(1) {
		ch = getchar();
		if(ch == EOF)
			break;
		if(ch == '<') {
			havetag = true;
			savechars.push(ch);
			ungetc(ch, stdin);
			break;
		}
		savechars.push(ch);
	}
	if(havetag) {
		while(1) {
			ch = getchar();
			if(ch == EOF)
				break;
			if(ch == '<') {
				savechars.push(ch);
				ch = getchar();
				if(ch != '/') {
					if(!escape)
						break;
					//cerr << endl << ch <<endl;
					TagType tagtype = getTagType(ch, '_');
					//cerr << endl << tagtype << " " << tt << endl;
					if(tagtype == tt) {
						found = true;
						savechars.push(ch);
						break;
					}
				}
				else {
					if(escape)
						break;
					break;
				}
			}
			/*else if(!isSpaceCharacter(ch)) {
				found = false;
				savechars.push(ch);
				//break;
			}*/
			savechars.push(ch);
		}
	}
	while(!savechars.empty()) {
		ch = savechars.top();
		ungetc(ch, stdin);
		savechars.pop();
	}
	//cerr << endl << "Found tag? " << found << " Looking for " << tt << endl;
	return found;
}


Optimizer::TagType Optimizer::getTagType(char l1, char l2)
{
	switch(l1) {
		case 'B': return Bold;
		case 'E':	return Emphasis;
		case 'I': return Italic;
		case 'P': return Plain;
		case 'S': return Strong;
		case 'T': return Typewriter;
		case 'U': return Underline;
		case '0': return Size0;
		case '1': return Size1;
		case '2': return Size2;
		case '3': return Size3;
		case '4': return Size4;
		case '5': return Size5;
		case '6': return Size6;
		case '7': return Size7;
		case '8': return Size8;
		case '9': return Size9;
		case 'r': return Red;
		case 'g': return Green;
		case 'b': return Blue;
		case 'c': return Cyan;
		case 'm': return Magenta;
		case 'y': return Yellow;
		case 'k': return Black;
		case 'w': return White;
	}
	cerr << "UNKOWN TAG: " << l1 << l2 << "!" << endl;
	return UnknownTag;
}

void Optimizer::printTag(TagType tt, bool escape)
{
	if(tt == Strong) {
		mStrong = !escape;
	}
	if(tt == Emphasis) {
		if(mStrong > 0)
			return;
	}
	if(tt == Typewriter) {
		if(!escape) {
			if(mTypewriter == 0)
				mTypewriter = mDoc;
			if(mTypewriter < mDoc)
				return;
		}
		else {
			//cerr << endl << "mBold: " << mBold << " mDoc: " << mDoc << endl;
			if(mTypewriter == 0)
				return;
			else if(mTypewriter == mDoc)
				mTypewriter = 0;
			else if(mTypewriter > mDoc)
				mTypewriter = 0;
			else if(mTypewriter < mDoc)
				return;
		}
	}
	if(tt == Bold) {
		if(!escape) {
			if(mBold == 0)
				mBold = mDoc;
			if(mBold < mDoc)
				return;
		}
		else {
			//cerr << endl << "mBold: " << mBold << " mDoc: " << mDoc << endl;
			if(mBold == 0)
				return;
			else if(mBold == mDoc)
				mBold = 0;
			else if(mBold > mDoc)
				mBold = 0;
			else if(mBold < mDoc)
				return;
		}
	}
	if(tt == Italic) {
		if(!escape) {
			if(mItalic == 0)
				mItalic = mDoc;
			if(mItalic < mDoc)
				return;
		}
		else {
			//cerr << endl << "mBold: " << mBold << " mDoc: " << mDoc << endl;
			if(mItalic == 0)
				return;
			else if(mItalic == mDoc)
				mItalic = 0;
			else if(mItalic > mDoc)
				mItalic = 0;
			else if(mItalic < mDoc)
				return;
		}
	}
	/*if(tt == Italic) {
		//if(mItalic != escape)
			mItalic = !escape;
		//else
		//	return;
	}
	*/
	
	putchar('<');
	if(escape)
		putchar('/');
	switch(tt) {
		case Bold: putchar('B'); break;
		case Emphasis: putchar('E'); putchar('M'); break;
		case Italic: putchar('I'); break;
		case Plain: putchar('P'); putchar('L'); break;
		case Strong: putchar('S'); break;
		case Typewriter: putchar('T'); putchar('T'); break;
		case Underline: putchar('U'); break;
		case Size0: putchar('0'); break;
		case Size1: putchar('1'); break;
		case Size2: putchar('2'); break;
		case Size3: putchar('3'); break;
		case Size4: putchar('4'); break;
		case Size5: putchar('5'); break;
		case Size6: putchar('6'); break;
		case Size7: putchar('7'); break;
		case Size8: putchar('8'); break;
		case Size9: putchar('9'); break;
		case Red: putchar('r'); break;
		case Green: putchar('g'); break;
		case Blue: putchar('b'); break;
		case Cyan: putchar('c'); break;
		case Magenta: putchar('m'); break;
		case Yellow: putchar('y'); break;
		case Black: putchar('k'); break;
		case White: putchar('w'); break;
	}
	putchar('>');
}

bool Optimizer::emElimination(bool escape)
{
	//putchar('$');
	
	return false;
}
